# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Matches BigBrain intesity profiles extracted by Wagstyl et al. to volumes."""

from . import query

from ..features.tabular import bigbrain_intensity_profile, layerwise_bigbrain_intensities
from ..features import anchor as _anchor
from ..commons import logger
from ..locations import point, pointset
from ..core import structure
from ..retrieval import requests, cache

import numpy as np
from typing import List
from os import path


class WagstylProfileLoader:

    REPO = "https://github.com/kwagstyl/cortical_layers_tutorial/raw/main"
    PROFILES_FILE_LEFT = "data/profiles_left.npy"
    THICKNESSES_FILE_LEFT = "data/thicknesses_left.npy"
    MESH_FILE_LEFT = "data/gray_left_327680.surf.gii"
    _profiles = None
    _vertices = None
    _boundary_depths = None

    def __init__(self):
        if self._profiles is None:
            self.__class__._load()

    @property
    def profile_labels(self):
        return np.arange(0., 1., 1. / self._profiles.shape[1])

    @classmethod
    def _load(cls):
        # read thicknesses, in mm, and normalize by their last column which is the total thickness
        thickness = requests.HttpRequest(f"{cls.REPO}/{cls.THICKNESSES_FILE_LEFT}").data.T
        total_thickness = thickness[:, :-1].sum(1)  # last column is the computed total thickness
        valid = np.where(total_thickness > 0)[0]
        cls._boundary_depths = np.c_[np.zeros_like(valid), (thickness[valid, :] / total_thickness[valid, None]).cumsum(1)]
        cls._boundary_depths[:, -1] = 1  # account for float calculation errors

        # find profiles with valid thickness
        profile_l_url = f"{cls.REPO}/{cls.PROFILES_FILE_LEFT}"
        if not path.exists(cache.CACHE.build_filename(profile_l_url)):
            logger.info(
                "First request to BigBrain profiles. Preprocessing the data "
                "now. This may take a little."
            )
        profiles_l_all = requests.HttpRequest(profile_l_url).data
        cls._profiles = profiles_l_all[valid, :]

        # read mesh vertices
        mesh_left = requests.HttpRequest(f"{cls.REPO}/{cls.MESH_FILE_LEFT}").data
        mesh_vertices = mesh_left.darrays[0].data
        cls._vertices = mesh_vertices[valid, :]

        logger.debug(f"{cls._profiles.shape[0]} BigBrain intensity profiles.")
        assert cls._vertices.shape[0] == cls._profiles.shape[0]

    def __len__(self):
        return self._vertices.shape[0]


cache.Warmup.register_warmup_fn()(lambda: WagstylProfileLoader._load())


class BigBrainProfileQuery(query.LiveQuery, args=[], FeatureType=bigbrain_intensity_profile.BigBrainIntensityProfile):

    def __init__(self):
        query.LiveQuery.__init__(self)

    def query(self, concept: structure.BrainStructure, **kwargs) -> List[bigbrain_intensity_profile.BigBrainIntensityProfile]:
        loader = WagstylProfileLoader()
        mesh_vertices = pointset.PointSet(loader._vertices, space='bigbrain')
        matched = concept.intersection(mesh_vertices)  # returns a reduced PointSet with og indices as labels
        if matched is None:
            return []
        assert isinstance(matched, pointset.PointSet)
        indices = matched.labels
        assert indices is not None
        features = []
        for i in matched.labels:
            anchor = _anchor.AnatomicalAnchor(
                location=point.Point(loader._vertices[i], space='bigbrain'),
                region=str(concept),
                species='Homo sapiens'
            )
            prof = bigbrain_intensity_profile.BigBrainIntensityProfile(
                anchor=anchor,
                depths=loader.profile_labels,
                values=loader._profiles[i],
                boundaries=loader._boundary_depths[i]
            )
            prof.anchor._assignments[concept] = _anchor.AnatomicalAssignment(
                query_structure=concept,
                assigned_structure=concept,
                qualification=_anchor.Qualification.CONTAINED,
                explanation=f"Surface vertex of BigBrain cortical profile was filtered using {concept}"
            )
            features.append(prof)

        return features


class LayerwiseBigBrainIntensityQuery(query.LiveQuery, args=[], FeatureType=layerwise_bigbrain_intensities.LayerwiseBigBrainIntensities):

    def __init__(self):
        query.LiveQuery.__init__(self)

    def query(self, concept: structure.BrainStructure, **kwargs) -> List[layerwise_bigbrain_intensities.LayerwiseBigBrainIntensities]:

        loader = WagstylProfileLoader()
        mesh_vertices = pointset.PointSet(loader._vertices, space='bigbrain')
        matched = concept.intersection(mesh_vertices)  # returns a reduced PointSet with og indices as labels
        if matched is None:
            return []
        assert isinstance(matched, pointset.PointSet)
        indices = matched.labels
        assert indices is not None
        matched_profiles = loader._profiles[indices, :]
        boundary_depths = loader._boundary_depths[indices, :]
        # compute array of layer labels for all coefficients in profiles_left
        N = matched_profiles.shape[1]
        prange = np.arange(N)
        layer_labels = 7 - np.array([
            [np.array([[(prange < T) * 1] for i, T in enumerate((b * N).astype('int'))]).squeeze().sum(0)]
            for b in boundary_depths
        ]).reshape((-1, 200))

        anchor = _anchor.AnatomicalAnchor(
            location=pointset.PointSet(loader._vertices[indices, :], space='bigbrain'),
            region=str(concept),
            species='Homo sapiens'
        )
        result = layerwise_bigbrain_intensities.LayerwiseBigBrainIntensities(
            anchor=anchor,
            means=[matched_profiles[layer_labels == layer].mean() for layer in range(1, 7)],
            stds=[matched_profiles[layer_labels == layer].std() for layer in range(1, 7)],
        )
        result.anchor._assignments[concept] = _anchor.AnatomicalAssignment(
            query_structure=concept,
            assigned_structure=concept,
            qualification=_anchor.Qualification.CONTAINED,
            explanation=f"Surface vertices of BigBrain cortical profiles were filtered using {concept}"
        )

        return [result]



# Copyright 2018-2024
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import contour
from ..locations import point
from ..core import parcellation

import numpy as np


class CorticalProfileSampler:
    """Samples cortical profiles from the cortical layer maps."""

    def __init__(self):
        self.layermap = parcellation.Parcellation.get_instance(
            "cortical layers"
        ).get_map(space="bigbrain", maptype="labelled")

    def query(self, query_point: point.Point):
        q = query_point.warp(self.layermap.space)
        smallest_dist = np.inf
        best_match = None
        for layername in self.layermap.regions:
            vertices = self.layermap.fetch(region=layername, format="mesh")["verts"]
            dists = np.sqrt(((vertices - q.coordinate) ** 2).sum(1))
            best = np.argmin(dists)
            if dists[best] < smallest_dist:
                best_match = (layername, best)
                smallest_dist = dists[best]

        best_vertex = best_match[1]
        hemisphere = "left" if "left" in best_match[0] else "right"
        print(f"Best match is vertex #{best_match[1]} in {best_match[0]}.")

        profile = [
            (_, self.layermap.fetch(region=_, format="mesh")["verts"][best_vertex])
            for _ in self.layermap.regions
            if hemisphere in _
        ]

        return contour.Contour(
            [p[1] for p in profile],
            space=self.layermap.space,
            labels=[p[0] for p in profile],
        )



# Copyright 2018-2024
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..volumes import volume
from ..locations import pointset, boundingbox
from ..commons import translation_matrix, y_rotation_matrix

import numpy as np
import math
from nilearn import image


class Patch:

    def __init__(self, corners: pointset.PointSet):
        """Construct a patch in physical coordinates.
        As of now, only patches aligned in the y plane of the physical space
        are supported."""
        # TODO: need to ensure that the points are planar, if more than 3
        assert len(corners) == 4
        assert len(np.unique(corners.coordinates[:, 1])) == 1
        self.corners = corners

    @property
    def space(self):
        return self.corners.space

    def flip(self):
        """Flips the patch. """
        self.corners._coordinates = self.corners.coordinates[[2, 3, 0, 1]]

    def extract_volume(self, image_volume: volume.Volume, resolution_mm: float):
        """
        fetches image data in a planar patch.
        TODO The current implementation only covers patches which are strictly
        define in the y plane. A future implementation should accept arbitrary
        oriented patches.accept arbitrary oriented patches.
        """
        assert image_volume.space == self.space

        # Extend the 2D patch into a 3D structure
        # this is only valid if the patch plane lies within the image canvas.
        canvas = image_volume.get_boundingbox()
        assert canvas.minpoint[1] <= self.corners.coordinates[0, 1]
        assert canvas.maxpoint[1] >= self.corners.coordinates[0, 1]
        XYZ = self.corners.coordinates
        voi = boundingbox.BoundingBox(
            XYZ.min(0)[:3], XYZ.max(0)[:3], space=image_volume.space
        )
        # enforce the patch to have the same y dimensions
        voi.minpoint[1] = canvas.minpoint[1]
        voi.maxpoint[1] = canvas.maxpoint[1]
        patch = image_volume.fetch(voi=voi, resolution_mm=resolution_mm)
        assert patch is not None

        # patch rotation defined in physical space
        vx, vy, vz = XYZ[1] - XYZ[0]
        alpha = -math.atan2(-vz, -vx)
        cx, cy, cz = XYZ.mean(0)
        rot_phys = np.linalg.multi_dot(
            [
                translation_matrix(cx, cy, cz),
                y_rotation_matrix(alpha),
                translation_matrix(-cx, -cy, -cz),
            ]
        )

        # rotate the patch in physical space
        affine_rot = np.dot(rot_phys, patch.affine)

        # crop in the rotated space
        pixels = (
            np.dot(np.linalg.inv(affine_rot), self.corners.homogeneous.T)
            .astype("int")
            .T
        )
        # keep a pixel distance to avoid black border pixels
        xmin, ymin, zmin = pixels.min(0)[:3] + 1
        xmax, ymax, zmax = pixels.max(0)[:3] - 1
        h, w = xmax - xmin, zmax - zmin
        affine = np.dot(affine_rot, translation_matrix(xmin, 0, zmin))
        return volume.from_nifti(
            image.resample_img(patch, target_affine=affine, target_shape=[h, 1, w]),
            space=image_volume.space,
            name=f"Rotated patch with corner points {self.corners} sampled from {image_volume.name}",
        )
